/*
	makethumb is copyright 2002 by Robert Kesterson, robertk@robertk.com,
	all rights reserved.  Portions based on the work on the Independent
	JPEG Group.
*/

#include <stdio.h>
#include <stdlib.h>
#include "makethumb.h"
#include <jpeglib.h>
#include <setjmp.h>
#include <string.h>


/*
	 override default error handler (which terminates the program...)
*/
	 
struct my_jpg_error_mgr 
{
  struct jpeg_error_mgr pub;	/* "public" fields */
  jmp_buf setjmp_buffer;		/* for return to caller */
  char *pWhat;					/* error message */
};

METHODDEF(void) my_jpg_error_handler(j_common_ptr cinfo)
{
  /* cinfo->err really points to a dtl_jpg_errpr_mgr struct, so coerce pointer */
  struct my_jpg_error_mgr *myerr = (struct my_jpg_error_mgr *) cinfo->err;
  /* save the error message */
  myerr->pWhat = (char *)strdup(cinfo->err->jpeg_message_table[cinfo->err->msg_code]);
  /* and return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}


/* allocate a new image of given size.  Must call FreeImage later. */
struct Image *CreateImage(long Width, long Height)
{
	struct Image *pImage;
	struct Pixel *pPixels;
	size_t nSize;

	pImage = NULL;
	nSize = Width * Height * sizeof(struct Pixel);
	pPixels =  (struct Pixel *)malloc(nSize);
	if (pPixels)
	{
		memset(pPixels, 0, nSize);
		pImage = (struct Image *)malloc(sizeof(struct Image));
		if (pImage)
		{
			pImage->Width = Width;
			pImage->Height = Height;
			pImage->pPixels = pPixels;
		}
		else
			free(pPixels);
	}
	return (pImage);
}

/* call this on any allocated image. */
void FreeImage(struct Image *pImage)
{
	if (pImage)
	{
		if (pImage->pPixels)
			free(pImage->pPixels);
		free (pImage);
	}
}

/* Read a JPEG image into memory.  Must call FreeImage later. */
struct Image *ReadJPEG(const char *pszFile)
{
	struct Image *pImage = NULL;
	struct jpeg_decompress_struct cinfo;
	struct my_jpg_error_mgr error_mgr;
	FILE *pFile = NULL;
	int nSignificantBits;
	int nRowWidth;
	int i, nRow, nCol, nOffset, nLines;
	JSAMPARRAY rowbuffer;

	error_mgr.pub.error_exit = my_jpg_error_handler;
	memset(&cinfo, 0, sizeof(struct jpeg_decompress_struct));
	cinfo.err = jpeg_std_error(&error_mgr.pub);
	jpeg_create_decompress(&cinfo);
	pFile = fopen(pszFile, "rb");
	if(!pFile)
	{
		fprintf(stderr, "Unable to open file");
		return NULL;
	}
	jpeg_stdio_src(&cinfo, pFile);
	if(setjmp(error_mgr.setjmp_buffer))
	{
		fprintf(stderr, error_mgr.pWhat);
		return NULL;
	}
	jpeg_read_header(&cinfo, 1);	/* "1" is to ignore table-only files */
	if(cinfo.out_color_space == JCS_GRAYSCALE)
		nSignificantBits = 8;
	cinfo.out_color_space = JCS_RGB;/* request RGB output in all cases */
	jpeg_start_decompress(&cinfo);

	pImage = CreateImage(cinfo.output_width, cinfo.output_height);
	nRowWidth = cinfo.output_width * cinfo.output_components;
	rowbuffer = (*cinfo.mem->alloc_sarray)(
					(j_common_ptr) &cinfo, JPOOL_IMAGE, nRowWidth, 8);
	if(!rowbuffer || !pImage)
	{
		fprintf(stderr, "Can't allocate memory");
		if(rowbuffer)
			free(rowbuffer);
		if(pImage)
			free(pImage);
		return NULL;
	}
	
	i = nOffset = 0;
	while(i < pImage->Height)
	{
		nLines = jpeg_read_scanlines(&cinfo, rowbuffer, 8);
		if(!nLines)
			break;
		for(nRow = 0; nRow < nLines; nRow++)
		{
			JSAMPLE *pSamples = rowbuffer[nRow];
			for(nCol = 0; nCol < pImage->Width; nCol++, nOffset++)
			{
				pImage->pPixels[nOffset].Red = GETJSAMPLE(pSamples[nCol * 3 + 0]);
				pImage->pPixels[nOffset].Green = GETJSAMPLE(pSamples[nCol * 3 + 1]);
				pImage->pPixels[nOffset].Blue = GETJSAMPLE(pSamples[nCol * 3 + 2]);
			}
		}
		i += nLines;
	}
	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);
	fclose(pFile);

	return(pImage);
}


/* 
	Create a scaled copy of an image.  Must call FreeImage on the copy later.
	Uses a DDA scaling algorithm, which gives a good image at a decent speed,
	whether scaling up or down.
*/
struct Image *ScaleImage(const struct Image *pImage, int newHeight)
{
	struct Image *pScaled;
	int newWidth;
    int x, y;
    int ddax, dday, izoom, i, j, k;
	long nMaxOffset, nMaxDiff;
	int wide, high, offset, offset2;
	double zoom;

	newWidth = ((float)newHeight / (float)pImage->Height) * (float)pImage->Width;
	pScaled = CreateImage(newWidth, newHeight);
	if(!pScaled)
	{
		fprintf(stderr, "Unable to allocate memory");
		return NULL;
	}

	nMaxOffset = newWidth * newHeight;
	wide = pImage->Width;
	high = pImage->Height;

	if(wide == newWidth && high == newHeight)
	{		// if same size, just copy it over and scram
		memcpy(pScaled->pPixels, pImage->pPixels, sizeof(struct Pixel) * wide * high);
		return;
	}
	/* "white out" the destination */
	memset(pScaled->pPixels, 0xFF, newWidth * newHeight * sizeof(struct Pixel));
    /* find differential amount */
	zoom = (double)newHeight / (high ? (double) high : 1);
    izoom = (int)(1.0/zoom * 1000);
    y = 0;	/* set destination row and vertical differential */
    dday = 0;
	offset = 0;
	/* Loop over rows in the original image */
    for (i = 0; i < high && offset < nMaxOffset; i++) 
	{
        /* set x offset and differential, set accumulated differential */
        dday -= 1000;
        x    = 0;
        ddax = 0;
        /* walk the original pixel row */
		nMaxDiff = nMaxOffset - offset;
        for (j = 0; j < wide; j++) 
		{            
            ddax -= 1000;	// Adjust the horizontal accumulated differential 
            while (ddax < 0) 
			{
		        /* Store values from original image scanline into the scaled */
			    /* buffer until accumulated differential crosses threshold */
				if(x < nMaxDiff)
				{
					offset2 = i * wide + j;
					pScaled->pPixels[offset + x].Red = pImage->pPixels[offset2].Red;
					pScaled->pPixels[offset + x].Green = pImage->pPixels[offset2].Green;
					pScaled->pPixels[offset + x].Blue = pImage->pPixels[offset2].Blue;
				}
                x++;
	            ddax += izoom;
		    }
        }
        while (dday < 0) 
		{
			/* now output the resized scanline repeatedly */
			/* until the vertical threshold is crossed	*/
            dday += izoom;	
			if( y * newWidth + x < nMaxOffset)	/* check for potential overflow */
				for (k = 0; (k <= x); k++) 
					pScaled->pPixels[(y * newWidth) + k] = pScaled->pPixels[offset + k]; 
            y++;		
        }
    }
	return(pScaled);
}

/* returns 0 on failure, 1 on success */
int WriteJPEG(const struct Image *pImage, const char *pszFile, int quality)
{
	struct jpeg_compress_struct cinfo;
	struct my_jpg_error_mgr error_mgr;
	FILE *pOutfile;		/* target file */
	JSAMPROW row_pointer[1];	/* pointer to JSAMPLE row[s] */
	int row_stride;		/* physical row width in image buffer */
	JSAMPLE *pSampleRow;
	int i;

	pSampleRow = (JSAMPLE *)malloc(pImage->Width * sizeof(JSAMPLE) * 3);
	cinfo.err = jpeg_std_error(&error_mgr.pub);
	jpeg_create_compress(&cinfo);
	pOutfile = fopen(pszFile, "wb");
	if (!pOutfile)
	{
		fprintf(stderr, "Can't open %s for writing\n", pszFile);
		return (0);
	}
	jpeg_stdio_dest(&cinfo, pOutfile);
	cinfo.image_width = pImage->Width;
	cinfo.image_height = pImage->Height;
	cinfo.input_components = 3;		/* # of color components per pixel */
	cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, TRUE /* limit to baseline-JPEG values */);
	jpeg_start_compress(&cinfo, TRUE);
	row_stride = pImage->Width * 3;	/* JSAMPLEs per row in image_buffer */
	row_pointer[0] = pSampleRow;	/* one-element array of scanlines */
	while (cinfo.next_scanline < cinfo.image_height) 
	{
		for (i = 0; i < pImage->Width; i++)
		{
			pSampleRow[(i * 3) + 0] = pImage->pPixels[cinfo.next_scanline * pImage->Width + i].Red;
			pSampleRow[(i * 3) + 1] = pImage->pPixels[cinfo.next_scanline * pImage->Width + i].Green;
			pSampleRow[(i * 3) + 2] = pImage->pPixels[cinfo.next_scanline * pImage->Width + i].Blue;
		}
	  (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}
	jpeg_finish_compress(&cinfo);
	fclose(pOutfile);
	jpeg_destroy_compress(&cinfo);
	free(pSampleRow);
	return 1;
}


int main(int argc, char **argv)
{
	int rc;
	struct Image *pImage, *pScaled;

	rc = 20;
	if (argc < 4)
	{
		printf("Usage is %s infile outfile thumbheight\n", argv[0]);
		return(10);
	}
	pImage = ReadJPEG(argv[1]);
	if (pImage)
	{
		pScaled = ScaleImage(pImage, atol(argv[3]));
		if (pScaled)
		{
			if (WriteJPEG(pScaled, argv[2], 85))
				rc = 0;
			FreeImage(pScaled);
		}
		FreeImage(pImage);
	}
	return(rc);
}
